#include "includes.h"

model_t sampleModel;
char sysBuffer[128];

int mmReadCh (FILE **file)
{
	return fgetc(*file);
}

void mmSkipLine (FILE **file)
{
	fgets(sysBuffer, 128, *file);
}

// mmReadNextParam: reads the next parameter from a *.m file stream,
// skips all unneeded characters
float mmReadNextParam (FILE **file)
{
	int ch = 0;
	float f = 0;
	unsigned char legal = 0;

	while (1)
	{
		ch = mmReadCh(file);
		
		if (ch == EOF)	{	// EOF or error: set err to 1 and break out of the loop
		 	printf("mmReadNextParam: unexpected EOF or parsing error\r\ncheck your *.m model file syntax\n");   
			exit(-1);
		}

		if ((ch >= 0x30  && ch <= 0x39) || ch == 0x2d) legal = 1;
		else legal = 0;
//		((ch >= 0x30  &&	// skip spaces or any other misc. characters (, . etc)
//		  ch <= 0x39) ||
//		  ch == 0x2d) ?		// a vertex param. could be negative so look for '-' (0x2d) while parsing
//		  legal = 1 : legal = 0;

		if (legal)
		{
			// back up by one character
			if (fseek(*file, (ftell(*file)- 1), SEEK_SET) != 0) {
				printf ("mmReadNextParam: fseek failed, sorry; check your *.m model file syntax\n");
				exit(-1);
			}

			// read the floating-point parameter and return it
			fscanf(*file, "%f", &f);
			return f;
		}
	}

	// EOF or error
	return 0;
}

// LoadModel(); loads an m file format model into model_t *model
void LoadModel (char *sFilename, model_t *model)
{
	FILE *file = fopen(sFilename, "r");
	int i, j;

	if (file) {
		// find the "beginmodel" tag
		while (1)
			if (strncmp(fgets(sysBuffer, 128, file), "beginmodel", 10) == 0)
				break;

		// read the number of polygons in this model
		sscanf(sysBuffer, "beginmodel %d", &model->nPolygons);

		// allocate enough memory for this model
		model->pList = (polygon_t *)malloc(model->nPolygons * sizeof(polygon_t));

		// read all polygon vertices
		for (i=0; i < model->nPolygons ; i++) {
			model->pList[i].v[0].x = mmReadNextParam(&file);
			model->pList[i].v[0].y = mmReadNextParam(&file);
			model->pList[i].v[0].z = mmReadNextParam(&file);
			model->pList[i].v[1].x = mmReadNextParam(&file);
			model->pList[i].v[1].y = mmReadNextParam(&file);
			model->pList[i].v[1].z = mmReadNextParam(&file);
			model->pList[i].v[2].x = mmReadNextParam(&file);
			model->pList[i].v[2].y = mmReadNextParam(&file);
			model->pList[i].v[2].z = mmReadNextParam(&file);
		}

		//calculate normal vectors
		for (i=0; i < model->nPolygons; i++) {
			normal_f( model->pList[i].v[0].x, model->pList[i].v[0].y, model->pList[i].v[0].z,
				model->pList[i].v[1].x, model->pList[i].v[1].y, model->pList[i].v[1].z,
				model->pList[i].v[2].x, model->pList[i].v[2].y, model->pList[i].v[2].z,		  
				&model->pList[i].n );
		}

		fclose(file);
	}
	else {
		printf("LoadModelM(): Cannot open file \"%s\"\n", sFilename);
		exit(-1);
	}
}

// DrawPolygon(); draws an opengl polygon defined by polygon_t
void DrawPolygon (polygon_t *p)
{
	glBegin(GL_TRIANGLES);

	glColor3f(0.8f, 0.8f, 0.8f);
	glNormal3f(p->n.x, p->n.y, p->n.z);
	glVertex3f(p->v[0].x, p->v[0].y, p->v[0].z);
	glVertex3f(p->v[1].x, p->v[1].y, p->v[1].z);
	glVertex3f(p->v[2].x, p->v[2].y, p->v[2].z);

	glEnd();
}

// DrawModel(); draws a model
void DrawModel_v (model_t *model, vector_t *translate, float r_anglex, float r_angley, float r_anglez)
{
    	int i;
	//glLoadIdentity();
	glTranslatef(translate->x, translate->y, translate->z);
	glRotatef(r_anglex, 1.0f, 0.0f, 0.0f);
	glRotatef(r_angley, 0.0f, 1.0f, 0.0f);
	glRotatef(r_anglez, 0.0f, 0.0f, 1.0f);
	
	glBegin(GL_TRIANGLES);
        glColor3f(0.8f, 0.8f, 0.8f);
		
	for (i=0; i<model->nPolygons; i++) {
		glNormal3f(model->pList[i].n.x, model->pList[i].n.y, model->pList[i].n.z);
	        glVertex3f(model->pList[i].v[0].x, model->pList[i].v[0].y, model->pList[i].v[0].z);
	        glVertex3f(model->pList[i].v[1].x, model->pList[i].v[1].y, model->pList[i].v[1].z);
        	glVertex3f(model->pList[i].v[2].x, model->pList[i].v[2].y, model->pList[i].v[2].z);
	}
	
	glEnd();
}

void DrawModel_f (model_t *model, float trans_x, float trans_y, float trans_z, 
	        float r_anglex, float r_angley, float r_anglez)
{
	vector_t translation;
	
	translation.x = trans_x;
	translation.y = trans_y;
	translation.z = trans_z;

	DrawModel_v (model, &translation, r_anglex, r_angley, r_anglez);
}

GLuint createModelDisplayList(model_t *model) {
	GLuint modelDL;
	
	modelDL = glGenLists(1);
	glNewList(modelDL,GL_COMPILE);
	
	DrawModel_f (model, 0, 0, 0, 0, 0, 0);

	glEndList();

	return (modelDL);
}

// FreeModel(); frees model memory
void FreeModel (model_t *model)
{
	if (model->pList)
	{
		free(model->pList);
		model->pList = NULL;
	}
}
