// simple.c - works with simple.world
// compile with:
//    gcc -Wall -g3 `pkg-config --cflags gazebo` -c -o simple.o simple.c
//    gcc simple.o -o simple `pkg-config --libs gazebo`
// Thanks to Kevin Knoedler for various fixes.

#include <stdio.h>
#include <gazebo.h>
#include <rtk.h>

#define COLOR_GRID_MINOR         0xE0E0E0
#define COLOR_GRID_MAJOR         0xC0C0C0

#define OBJECTS	4

struct {
    char *name;
    double sx, sy;
    int rgb32;
} object[4] = {
    {"watcher_truth", 0.50, 0.36, 0x843000 },
    {"pusher1_truth", 0.50, 0.36, 0xFF0000},
    {"pusher2_truth", 0.50, 0.36, 0x00FF00 },
    {"puck_truth", 2.0, 0.5, 0x0000FF }
};

int main (int argc, char *argv[])
{
  int i;
  int server_id;
  int client_id;
  gz_client_t *client;
  gz_truth_t *truth[OBJECTS];

  rtk_app_t *app;
  rtk_canvas_t *canvas;
  rtk_fig_t *root_fig;
  rtk_fig_t *base_fig;
  rtk_fig_t *grid_fig;

  rtk_init(&argc, &argv);

  app = rtk_app_create();
  canvas = rtk_canvas_create(app);
  rtk_canvas_title(canvas, "Gazebo 2D GUI");

  rtk_canvas_size(canvas, 600, 600);
  rtk_canvas_origin(canvas, 0, 0);
  rtk_canvas_scale(canvas, 0.02, 0.02);

  root_fig = rtk_fig_create(canvas, NULL, 0);
  base_fig = rtk_fig_create(canvas, NULL, 0);
  grid_fig = rtk_fig_create(canvas, NULL, -10);

  rtk_fig_grid(grid_fig, 0, 0, -1, -1, 2);

  rtk_fig_color_rgb32(grid_fig, COLOR_GRID_MAJOR);
  rtk_fig_grid(grid_fig, 0, 0, 51.2, 51.2, 1);

  rtk_app_main_init(app);

  // Assume one server and one client
  server_id = 0;
  client_id = 0;

  // Print all errors
  gz_error_init(1, 9);

  // Create a client object
  client = gz_client_alloc();

  // Connect to the server
  if (gz_client_connect_wait(client, server_id, client_id) != 0)
    return -1;

  // Create a position interface
  for (i = 0; i < OBJECTS; i++) {
    truth[i] = gz_truth_alloc();
    if (gz_truth_open(truth[i], client, object[i].name) != 0)
	return -1;
  }

  // Connect to device on server

  rtk_canvas_render(canvas);

  // Read some data
  for (;;)
  {
    // Wait for new data
    gz_client_wait(client);

    rtk_fig_clear(base_fig);

    // Lock, Draw, Unlock
    for (i = 0; i < OBJECTS; i++) {
      gz_truth_lock(truth[i], 1);
      rtk_fig_color_rgb32(base_fig, object[i].rgb32);
      rtk_fig_rectangle(base_fig,
	    truth[i]->data->pos[0], truth[i]->data->pos[1],
	    truth[i]->data->rot[2],
	    object[i].sx, object[i].sy,
	    1);
      /*
      printf("%d: [ %0.3f %0.f %0.3f ]\n", i,
	    truth[i]->data->pos[0], truth[i]->data->pos[1],
	    truth[i]->data->rot[2]);
      */
      gz_truth_unlock(truth[i]);
    }

    if (!rtk_app_main_loop(app)) break;

    rtk_canvas_render(canvas);
  }

  // Close the interface and free object
  for (i = 0; i < OBJECTS; i++) {
    gz_truth_close(truth[i]);
    gz_truth_free(truth[i]);
  }

  // Close the connection and free the client object
  gz_client_disconnect(client);
  gz_client_free(client);

  rtk_app_main_term(app);
  rtk_fig_destroy(base_fig);
  rtk_fig_destroy(root_fig);

  rtk_canvas_destroy(canvas);
  rtk_app_destroy(app);

  return 0;
}
