/*
 *  Player - One Hell of a Robot Server
 *  Copyright (C) 2003  
 *     Brian Gerkey, Andrew Howard
 *                      
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/*
 * Desc: Player driver for laser-stabilized odometry
 * Author: Andrew Howard
 * Date: 24 Nov 2004
 * CVS: $Id: lodo_driver.cc,v 1.2 2006/06/03 08:32:42 kniewiar Exp $
 */

/** @defgroup lodo_driver Laser-stabilized odometry (lodo_driver)

The lodo driver provides laser-stabilized odometric pose estimates:
the inputs are raw odometry and laser scans, the outputs are corrected
pose estimates and synchronized laser scans (i.e., no lag-lead
effects).  The drift in this corrected estimate is much, much, lower
than that seen with odometry alone (e.g., less than 1m drift in
position after 100m of travel).  The lodo driver is a thin wrapper
around a free-standing laser-odometry library (liblodo); see lodo.h
for details of how the library works.

To use the lodo driver, simply point it at a suitable pair of position
and laser devices.  Generally, the position device will be provided by
to a robot driver, such @ref player_driver_p2os, while the laser
device will be provided by the @ref player_driver_sicklms200 driver.
The lodo driver provides a new pair of position and laser interfaces
that provide corrected pose estimates and synchronized laser data,
respectively.  If you have a Pioneer-like robot with a SICK laser, the
lodo driver will be almost entirely transparent to your client
programs: simply get odometry and laser data from the lodo driver
instead of the underlying hardware drivers.

@par Laser configuration

For best performance, set the laser range and bearing resolutions
to 10cm and 1 degree, respectively; for the SICK, this corresponds to:

@verbatim
driver
(
  name "sicklms200"
  resolution 100
  range_res 10
  rate 38400
)
@endverbatim

It is also vitally important that the laser be rigidly mounted to
the robot (that means no velcro or gaffer tape), with a known, fixed
orientation.  Badly aligned lasers introduce systematic errors and high
drift rates (worse than odometry alone).

See the caveats below for additional usage notes.


@par Caveats

While the driver has been extensively tested on certain hardware
configurations (Pioneer2DX robots with SICK LMS200 lasers), mileage
may vary on other platforms. Several points should be noted:

- The incremental pose corrections are applied to orientation @e only; this
works great on robots with good linear odometry (like the Pioneer2DX),
but is likely to fail on a skid-steered Pioneer2AT (don't know,
haven't tried it).

- The driver has been tested with a SICK LMS200 operating at 38400 baud,
1 degree angular resolution and 10cm range resolution (yields 181 samples at
10Hz).  Higher data rates (e.g., 75Hz are not recommended unless you have
a very fast processor).

- The library is fairly compute intensive; on a 2.6 GHz P4, processing
time for each scan is of the order of 10ms.



@par Requires

- @ref player_interface_position
- @ref player_interface_laser


@par Provides

- @ref player_interface_position
- @ref player_interface_laser


@par Supported configuration requests

None

@todo Pass configuration requests through to underlying drivers.


@par Configuration file options

- plugin (filename)
  - Default: "lodo_driver.so"
  - Loads the driver as a plugin.

- laser_pose (float tuple)
  - Default: [0 0 0]
  - Pose of the laser sensor in the robot's coordinate system (m, m, degrees).

- max_range (float length)
  - Default: 8.0
  - Maximum valid laser range.
  

@todo Expose library options.
  
@par Example 

@verbatim
device
(
  plugin "lodo_driver.so"
  driver "lodo_driver"
  requires ["position:0" "laser:0"]
  provides ["position:10" "laser:10"]
)
@endverbatim

@par Authors

Andrew Howard

*/

#include <unistd.h>
#include <string.h>
#include <netinet/in.h>

#include <player/drivertable.h>
#include <player/devicetable.h>
#include <player/driver.h>
#include <player/error.h>
#include <player.h>

#include <playerc.h>

#include "lodo.h"


// Use this to replace missing declaration in player 1.6
extern "C" void playerc_position_putdata(playerc_position_t *device, player_msghdr_t *header,
                                         player_position_data_t *data, size_t len);
extern "C" void playerc_ptz_putdata(playerc_ptz_t *device, player_msghdr_t *header,
                                         player_ptz_data_t *data, size_t len);


////////////////////////////////////////////////////////////////////////////////
/// @brief Driver class for laser-stabilized odometry (lodo)
class LodoDriver : public Driver
{
  /// @brief Constructor; need that
  public: LodoDriver(ConfigFile* cf, int section);

  /// @brief Initialize the driver
  public: int Setup();

  /// @brief Finalize the driver
  public: int Shutdown();

  /// @brief Main function for device thread.
  private: virtual void Main();

  /// @brief Update lodo estimate
  private: bool UpdateLodo();

  private: bool PTZ_Angle_OK();
	   
  /// @brief Process pending configuraiton requests
  private: void CheckConfig();

  /// @brief Check for new commands
  private: void CheckCommands();

  /// @brief Update the outgoing data
  private: void RefreshData();

  /// Laser offset in robot cs
  private: pose2_t laser_pose;

  /// Max valid laser range
  private: double max_range;
  
  /// Required position interface
  private: player_device_id_t position_id;
  private: Driver *position_driver;
  private: player_position_data_t position_data;
  private: playerc_position_t *position;

  /// Required ptz interface
  private: player_device_id_t ptz_id;
  private: Driver *ptz_driver;
  private: player_ptz_data_t ptz_data;
  private: playerc_ptz_t *ptz;

  /// Required laser interface
  private: player_device_id_t laser_id;
  private: Driver *laser_driver;
  private: player_laser_data_t laser_data;    
  private: playerc_laser_t *laser;
  private: struct timeval laser_time;
  private: int laser_data_size;
  
  /// Provided position interface
  private: player_device_id_t out_position_id;
  private: player_position_data_t out_position_data;
  private: player_position_cmd_t out_position_cmd;

  /// Provided laser interface
  private: player_device_id_t out_laser_id;

  /// Lodo lib handler
  private: lodo_t *lodo;

  /// Corrected robot pose
  private: pose2_t robot_pose;
};


// A factory creation function, declared outside of the class so that it
// can be invoked without any object context (alternatively, you can
// declare it static in the class).  In this function, we create and return
// (as a generic Driver*) a pointer to a new instance of this driver.
Driver* LodoDriver_Init(ConfigFile* cf, int section)
{
  // Create and return a new instance of this driver
  return ((Driver*) (new LodoDriver(cf, section)));
}

// A driver registration function, again declared outside of the class so
// that it can be invoked without object context.  In this function, we add
// the driver into the given driver table, indicating which interface the
// driver can support and how to create a driver instance.
void LodoDriver_Register(DriverTable* table)
{
  table->AddDriver("lodo_driver", LodoDriver_Init);
}


////////////////////////////////////////////////////////////////////////////////
// Extra stuff for building a shared object.

// need the extern to avoid C++ name-mangling
extern "C"
{
  int player_driver_init(DriverTable* table)
  {
    puts("plugin init");
    LodoDriver_Register(table);
    return(0);
  }
}


////////////////////////////////////////////////////////////////////////////////
// Constructor.  Retrieve options from the configuration file and do any
// pre-Setup() setup.
LodoDriver::LodoDriver(ConfigFile* cf, int section)
    : Driver(cf, section)
{

  // Get incoming position interface
  if (cf->ReadDeviceId(&(this->position_id), section, "requires",
                       PLAYER_POSITION_CODE, -1, NULL) != 0)
  {
    this->SetError(-1);
    return;
  }

  // Get incoming ptz interface
  if (cf->ReadDeviceId(&(this->ptz_id), section, "requires",
                       PLAYER_PTZ_CODE, -1, NULL) != 0)
  {
    this->SetError(-1);
    return;
  }

  // Get incoming laser interface
  if (cf->ReadDeviceId(&(this->laser_id), section, "requires",
                       PLAYER_LASER_CODE, -1, NULL) != 0)
  {
    this->SetError(-1);
    return;
  }

  // Create position interface
  if (cf->ReadDeviceId(&(this->out_position_id), section, "provides",
                       PLAYER_POSITION_CODE, -1, NULL) != 0)
  {
    this->SetError(-1);
    return;
  }  
  if (this->AddInterface(this->out_position_id, PLAYER_ALL_MODE,
                         sizeof(player_position_data_t),
                         sizeof(player_position_cmd_t), 10, 10) != 0)
  {
    this->SetError(-1);    
    return;
  }

  // Create laser interface
  if (cf->ReadDeviceId(&(this->out_laser_id), section, "provides",
                       PLAYER_LASER_CODE, -1, NULL) != 0)
  {
    this->SetError(-1);
    return;
  }    
  if (this->AddInterface(this->out_laser_id, PLAYER_READ_MODE,
                         sizeof(player_laser_data_t), 0, 10, 10) != 0)
  {
    this->SetError(-1);        
    return;
  }

  // Laser offset from robot origin
  this->laser_pose.pos.x = cf->ReadTupleLength(section, "laser_pose", 0, 0.0);
  this->laser_pose.pos.x = cf->ReadTupleLength(section, "laser_pose", 1, 0.0);
  this->laser_pose.rot = cf->ReadTupleAngle(section, "laser_pose", 2, 0.0);

  // Max valid range
  this->max_range = cf->ReadLength(section, "max_range", 8.00);

  return;
}


////////////////////////////////////////////////////////////////////////////////
// Set up the device.  Return 0 if things go well, and -1 otherwise.
int LodoDriver::Setup()
{   
  // Subscribe to the position
  this->position_driver = deviceTable->GetDriver(this->position_id);
  if (!this->position_driver)
  {
    PLAYER_ERROR("unable to locate suitable position device");
    return -1;
  }
  if (this->position_driver->Subscribe(this->position_id) != 0)
  {
    PLAYER_ERROR("unable to subscribe to position device");
    return -1;
  }

  // Subscribe to the laser
  this->laser_driver = deviceTable->GetDriver(this->laser_id);
  if (!this->laser_driver)
  {
    PLAYER_ERROR("unable to locate suitable laser device");
    return -1;
  }
  if (this->laser_driver->Subscribe(this->laser_id) != 0)
  {
    PLAYER_ERROR("unable to subscribe to laser device");
    return -1;
  }

  // Subscribe to the ptz
  this->ptz_driver = deviceTable->GetDriver(this->ptz_id);
  if (!this->ptz_driver)
  {
    PLAYER_ERROR("unable to locate suitable ptz device");
    return -1;
  }
  if (this->ptz_driver->Subscribe(this->ptz_id) != 0)
  {
    PLAYER_ERROR("unable to subscribe to ptz device");
    return -1;
  }
  
  // Create proxies
  this->position = playerc_position_create(NULL, -1);
  assert(this->position);
  this->laser = playerc_laser_create(NULL, -1);
  assert(this->laser);
  this->ptz = playerc_ptz_create(NULL, -1);
  assert(this->ptz);

  // Defer creation of lodo object so we can auto-dectect the
  // laser settings
  this->lodo = NULL;
      
  // Start the device thread; spawns a new thread and executes
  // LodoDriver::Main(), which contains the main loop for the driver.
  this->StartThread();

  return(0);
}


////////////////////////////////////////////////////////////////////////////////
// Shutdown the device
int LodoDriver::Shutdown()
{
  // Stop and join the driver thread
  this->StopThread();

  // Free lodo
  if (this->lodo)
    lodo_free(this->lodo);
  
  // Free proxies
  playerc_position_destroy(this->position);
  playerc_laser_destroy(this->laser);
  
  // Release required drivers
  this->laser_driver->Unsubscribe(this->laser_id);
  this->position_driver->Unsubscribe(this->position_id);
  this->ptz_driver->Unsubscribe(this->ptz_id);
  
  return(0);
}


////////////////////////////////////////////////////////////////////////////////
// Main function for device thread
void LodoDriver::Main() 
{
  PLAYER_MSG0(1, "lodo is running");
  
  // The main loop; interact with the device here
  while (true)
  {
    // Test if we are supposed to cancel
    pthread_testcancel();

    // Wait for new laser data
    this->laser_driver->Wait();

    if (!this->PTZ_Angle_OK())
      continue;
    // Update estimates
    if (this->UpdateLodo())
      this->RefreshData();

    // Check for and handle configuration requests
    this->CheckConfig();

    // Check for commands
    this->CheckCommands();
  }
  return;
}


////////////////////////////////////////////////////////////////////////////////
// Get PTZ angle
bool LodoDriver::PTZ_Angle_OK()
{
  size_t size;
  struct timeval time;
  double real_tilt;
  player_ptz_data_t ptzdata;

  // Get latest odometry data
  size = this->ptz_driver->GetData(this->ptz_id, &this->ptz_data,
                                        sizeof(this->ptz_data), &time);
  if (size <= 0) 
    return false;

  memcpy(&ptzdata, &this->ptz_data, sizeof(ptzdata));

  // Convert to a useable format
  playerc_ptz_putdata(this->ptz, NULL, &ptzdata, size);
  real_tilt = this->ptz->tilt * 180.0 / M_PI;
  if (fabs(real_tilt) < 0.2)
    return true;
  fprintf (stderr, "TILT = %lf! Lodo will not work...\n", this->ptz->tilt * 180.0 / M_PI);  fflush(stderr);
  return false;
}


////////////////////////////////////////////////////////////////////////////////
// Update lodo estimate
bool LodoDriver::UpdateLodo()
{
  size_t size;
  struct timeval time;
  player_laser_data_t ldata;
  pose2_t pose;

  // Get latest odometry data
  size = this->position_driver->GetData(this->position_id, &this->position_data,
                                        sizeof(this->position_data), &time);
  if (size <= 0)
    return false;

  // Convert to a useable format
  playerc_position_putdata(this->position, NULL, &this->position_data, size);
  this->position->info.datatime = time.tv_sec + 1e-6 * time.tv_usec;

  // Get latest laser data
  size = this->laser_driver->GetData(this->laser_id, &this->laser_data,
                                     sizeof(this->laser_data), &time);
  if (size <= 0)
    return false;
  this->laser_time = time;
  this->laser_data_size = size;

  // Convert to a useable format
  // TODO: remove copy once the put data function is fixed (should
  // not modify the data argument, but it does).
  memcpy(&ldata, &this->laser_data, sizeof(ldata));
  playerc_laser_putdata(this->laser, NULL, &ldata, size);
  this->laser->info.datatime = time.tv_sec + 1e-6 * time.tv_usec;

  // Auto-detect range count and create lodo object
  if (this->lodo == NULL)
  {
    PLAYER_MSG1(2, "autodetected %d range readings", this->laser->scan_count);
    if (this->laser->scan_count != 181)
      PLAYER_WARN("range readings != 181; please set your laser in 1 degree resolution mode");

    this->lodo = lodo_alloc(this->laser->scan_count, this->max_range, 0.1,
                            this->laser->scan_start, this->laser->scan_res);
    this->lodo->laser_pose = this->laser_pose;
  }

  // Check for changes in the scan count
  if (this->laser->scan_count != this->lodo->num_ranges)
  {
    PLAYER_ERROR2("incorrect range count; expecting %d, got %d",
                  this->laser->scan_count, this->lodo->num_ranges);
    return false;
  }

  // Massage odometric pose
  pose.pos = vector2_set(this->position->px, this->position->py);
  pose.rot = this->position->pa;

  // Update and correct pose
  pose = lodo_add_scan(this->lodo, pose, this->laser->scan_count, this->laser->ranges);

  // Store corrected robot pose
  this->robot_pose = pose;

  // TESTING
//  fprintf(stderr, "%d %2.2f ODOM(X,Y,th) = (%.2f,%.2f,%.1f) LODO(X,Y,th) = (%.2f,%.2f,%.1f)\n", 
  fprintf(stderr, "%d ODOM(X,Y,th) = (%.2f,%.2f,%.1f) LODO(X,Y,th) = (%.2f,%.2f,%.1f)\n", 
	 lodo->scan_count, 
//	 lodo->fit_correct,
         lodo->scan.opose.pos.x, lodo->scan.opose.pos.y, lodo->scan.opose.rot * 180.0 / M_PI,
         lodo->scan.cpose.pos.x, lodo->scan.cpose.pos.y, lodo->scan.cpose.rot * 180.0 / M_PI);
  fflush(stderr);

  return true;
}


////////////////////////////////////////////////////////////////////////////////
// Process pending configuration requests
void LodoDriver::CheckConfig()
{
  void *client;
  unsigned char buffer[PLAYER_MAX_REQREP_SIZE];
  
  while (this->GetConfig(this->out_position_id, &client, &buffer, sizeof(buffer), NULL) > 0)
  {
    if (this->PutReply(this->out_position_id, client, PLAYER_MSGTYPE_RESP_NACK,
                       NULL, 0, NULL) != 0)
      PLAYER_ERROR("PutReply() failed");
  }

  while (this->GetConfig(this->out_laser_id, &client, &buffer, sizeof(buffer), NULL) > 0)
  {
    if (this->PutReply(this->out_laser_id, client, PLAYER_MSGTYPE_RESP_NACK,
                       NULL, 0, NULL) != 0)
      PLAYER_ERROR("PutReply() failed");
  }

  return;
}


////////////////////////////////////////////////////////////////////////////////
// Process new commands
void LodoDriver::CheckCommands()
{
  this->GetCommand(this->out_position_id, &this->out_position_cmd, sizeof(this->out_position_cmd), NULL);
  
  return;
}


////////////////////////////////////////////////////////////////////////////////
// Write new data
void LodoDriver::RefreshData()
{
  // Write position data; use the corrected pose and the laser timestamp
  memcpy(&this->out_position_data, &this->position_data, sizeof(this->out_position_data));
  this->out_position_data.xpos = htonl((int32_t) (int) (this->robot_pose.pos.x * 1000));
  this->out_position_data.ypos = htonl((int32_t) (int) (this->robot_pose.pos.y * 1000));
  this->out_position_data.yaw = htonl((int32_t) (int) (this->robot_pose.rot * 180 / M_PI));  
  this->PutData(this->out_position_id, &this->out_position_data,
                sizeof(this->out_position_data), &this->laser_time);
  
  // Push the incoming laser data to the outgoing interface
  this->PutData(this->out_laser_id, &this->laser_data,
                this->laser_data_size, &this->laser_time);

  return;
}
