/*
 *  Gazebo - Outdoor Multi-Robot Simulator
 *  Copyright (C) 2003  
 *     Nate Koenig & Andrew Howard
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/* Desc: Model for a Protonek
 * Author: Andrew Howard
 * Date: 8 May 2003
 * CVS: $Id: Protonek.cc,v 1.1.1.1 2006/04/18 20:49:22 kniewiar Exp $
 */

/// @addtogroup models 
/// @{
/** @defgroup Protonek Pioneer 2 AT

@htmlinclude Protonek_view.html

The Protonek model simulates an ActivMedia Protonek mobile robot
base with 16 sonars.

@par libgazebo interfaces

- Position information is available through the @ref position interface.
- Power information is available through the @ref power interface.
- Sonar information is available through the @ref sonar interface.


@par Player drivers

- Position information is available through the %gz_position driver.
- Power information is available through the %gz_power driver.
- Sonar information is available through the %gz_sonar driver.

@par Attributes

The following attributes are supported.

@htmlinclude default_attr_include.html

- updateRate (float, Hz)
  - Updates per second
  - Default: 10

- batteryLevel (float, volts)
  - Initial battery level
  - Default: 12.4

- batteryCurve (float)
  - Discharge curve: about one hour quiescent
  - Default: 2/3600  2/1e4


@par Bodies

The following bodies are created by this model.

@htmlinclude default_body_include.html

@par Example

@verbatim
<model:Protonek>
  <xyz>0 0 0</xyz>
</model:Protonek>
@endverbatim

@par Views

@htmlinclude Protonek_more_views.html

@par Authors

Andrew Howard, Nate Koenig

*/
/// @} 

#include <assert.h>
#include <gazebo.h>
#include <gazebo/World.hh>
#include <gazebo/WorldFile.hh>
#include <gazebo/ModelFactory.hh>
#include <gazebo/Body.hh>
#include <gazebo/BoxGeom.hh>
#include <gazebo/SphereGeom.hh>
#include <gazebo/WheelGeom.hh>
#include <gazebo/PrismGeom.hh>
#include <gazebo/HingeJoint.hh>
#include <gazebo/RayProximity.hh>

#include "Protonek.hh"



//////////////////////////////////////////////////////////////////////////////
// Register this model
GZ_REGISTER_PLUGIN("Protonek", Protonek)


//////////////////////////////////////////////////////////////////////////////
// Constructor
Protonek::Protonek( World *world )
    : Model( world )
{
  int i;
  this->chassis = NULL;
  for (i=0; i<6; i++)
  {
    this->wheels[i] = NULL;
    this->wheelJoints[i] = NULL;
  }

  return;
}


//////////////////////////////////////////////////////////////////////////////
// Destructor
Protonek::~Protonek()
{
  int i;
  if (this->chassis)
    delete this->chassis;

  this->chassis = NULL;
  for (i=0; i<6; i++)
  {
    if (this->wheels[i])
      delete this->wheels[i];

    if (this->wheelJoints[i])
      delete this->wheelJoints[i];

    this->wheels[i] = NULL;
    this->wheelJoints[i] = NULL;
  }


  return;
}


//////////////////////////////////////////////////////////////////////////////
// Load the model
int Protonek::Load( WorldFile *file, WorldFileNode *node )
{
  this->wheelSep = 0.335;
  this->wheelDiam = 0.10;

  this->updatePeriod = 1.0 / (node->GetDouble("updateRate", 10) + 1e-6);
  this->updateTime = -updatePeriod;

  // Create the ODE objects
  if (this->OdeLoad(file, node) != 0)
    return -1;

  // Create the sonar sensor
  if (this->LoadSonar(file, node) != 0)
    return -1;

  // Initial battery level
  this->batteryLevel = node->GetDouble("batteryLevel", 12.4);

  // Discharge curve: about one hour quiescent
  this->batteryCurve[0] = node->GetTupleDouble("batteryCurve", 0, 2 / 3600.0);
  this->batteryCurve[1] = node->GetTupleDouble("batteryCurve", 1, 2 / 1e4);  

  return 0;
}


//////////////////////////////////////////////////////////////////////////////
// Load ODE objects
int Protonek::OdeLoad( WorldFile *file, WorldFileNode *node )
{
  int i;
  double length, width, height, oz;
  double mass;
  double wheelSep, wheelDiam, wheelThick;
  Geom *geom;

  length = 0.50;
  width = 0.360;
  height = 0.095;
  mass = 10.0;

  wheelSep = this->wheelSep;
  wheelDiam = this->wheelDiam;
  wheelThick = 0.025;

  oz = wheelDiam / 2;
  
  // Create the main chassis of the robot
  this->chassis = new Body( this->world );

  geom = new BoxGeom( this->chassis, this->modelSpaceId, 0.38, 0.28, 0.075);
  geom->SetRelativePosition( GzVectorSet(0, 0, oz+0.03) );
  geom->SetMass( mass );
  geom->SetColor( GzColor(0.8, 0.78, 0.78) );  

  geom = new BoxGeom( this->chassis, this->modelSpaceId, length-2*0.04, width, 0.025);
  geom->SetRelativePosition( GzVectorSet(0, 0, oz+0.03+0.075/2+0.025/2) );
  geom->SetMass( 0 );
  geom->SetColor( GzColor(0.8, 0.78, 0.78) );  

  geom = new BoxGeom( this->chassis, this->modelSpaceId, 0.04, width, 0.05);
  geom->SetRelativePosition( GzVectorSet(length/2-0.02, 0, oz+0.03+0.075/2) );
  geom->SetMass( 0 );
  geom->SetColor( GzColor(0.8, 0.78, 0.78) );  

  geom = new BoxGeom( this->chassis, this->modelSpaceId, 0.04, width, 0.05);
  geom->SetRelativePosition( GzVectorSet(-length/2+0.02, 0, oz+0.03+0.075/2) );
  geom->SetMass( 0 );
  geom->SetColor( GzColor(0.8, 0.78, 0.78) );  

#if 1
  geom = new BoxGeom( this->chassis, this->modelSpaceId, 0.3, width-0.08, 0.08);
  geom->SetRelativePosition( GzVectorSet(-0.05, 0, 0.1825));
  geom->SetMass( 0 );
  geom->SetColor( GzColor(0.8, 0.78, 0.78) );  
#else
  // Bottom chassis
  double extractVector[3] = {0., width-0.08, 0.0};
  double vertices[12] = {0.0, 0.0, 0.0,
			 0.41, 0.0, 0.0,
			 0.41, 0.0, 0.08,
			 0.065, 0.0, 0.08};
  
  geom = new PrismGeom( this->chassis, this->modelSpaceId, 0, extractVector, 4, vertices );
  geom->SetRelativePosition( GzVectorSet(0, 0, 0.1825) );
  geom->SetMass( 0 );
  geom->SetColor( GzColor(1, 0, 0) );
#endif

  this->AddBody( this->chassis, true );

  // Create the wheels
  for (i = 0; i < 6; i++)
  {
    this->wheels[i] = new Body( this->world );

    geom = new WheelGeom( this->wheels[i], this->modelSpaceId, wheelDiam/2, wheelThick/2);
    geom->SetRelativePosition( GzVectorSet(0, 0, oz) );
    geom->SetMass( 1.0 );
    geom->SetColor( GzColor(0.3, 0.3, 0.3) );

    this->AddBody( this->wheels[i] );
  }

  this->wheels[0]->SetPosition(GzVectorSet(0.15, +0.5 * wheelSep - 2*wheelThick, oz));
  this->wheels[0]->SetRotation(GzQuaternFromAxis(1, 0, 0, -M_PI / 2));
  this->wheels[1]->SetPosition(GzVectorSet(0.15, -0.5 * wheelSep + 2*wheelThick, oz));
  this->wheels[1]->SetRotation(GzQuaternFromAxis(1, 0, 0, +M_PI / 2));
  this->wheels[2]->SetPosition(GzVectorSet(0.0, +0.5 * wheelSep - 2*wheelThick, oz));
  this->wheels[2]->SetRotation(GzQuaternFromAxis(1, 0, 0, -M_PI / 2));
  this->wheels[3]->SetPosition(GzVectorSet(0.0, -0.5 * wheelSep + 2*wheelThick, oz));
  this->wheels[3]->SetRotation(GzQuaternFromAxis(1, 0, 0, +M_PI / 2));
  this->wheels[4]->SetPosition(GzVectorSet(-0.15, +0.5 * wheelSep - 2*wheelThick, oz));
  this->wheels[4]->SetRotation(GzQuaternFromAxis(1, 0, 0, -M_PI / 2));
  this->wheels[5]->SetPosition(GzVectorSet(-0.15, -0.5 * wheelSep + 2*wheelThick, oz));
  this->wheels[5]->SetRotation(GzQuaternFromAxis(1, 0, 0, +M_PI / 2));
  
  // Attach the wheels to the chassis
  for (i = 0; i < 6; i++)
  {
    this->wheelJoints[i] = new HingeJoint( this->world->worldId );
    this->wheelJoints[i]->Attach( this->wheels[i], this->chassis );

    GzVector a = wheels[i]->GetPosition();    
    this->wheelJoints[i]->SetAnchor( a.x, a.y, a.z );
    this->wheelJoints[i]->SetAxis( 0, 1, 0 );
    this->wheelJoints[i]->SetParam( dParamSuspensionERP, 0.4 );
    this->wheelJoints[i]->SetParam( dParamSuspensionCFM, 0.8 );
  }

  return 0;
}


//////////////////////////////////////////////////////////////////////////////
// Load the sonar
int Protonek::LoadSonar( WorldFile *file, WorldFileNode *node )
{
  int i;
  GzVector a, b, p;
  p.z = 0.095;
  
  // TODO: check dimensions
    
  pos[0]  = GzVectorSet( 0.25, +0.15, p.z );
  pos[1]  = GzVectorSet( 0.25, +0.12, p.z );
  pos[2]  = GzVectorSet( 0.25, +0.09, p.z );
  pos[3]  = GzVectorSet( 0.25, +0.06, p.z );
  pos[4]  = GzVectorSet( 0.25, +0.03, p.z );
  pos[5]  = GzVectorSet( 0.25, +0.00, p.z );
  pos[6]  = GzVectorSet( 0.25, -0.03, p.z );
  pos[7]  = GzVectorSet( 0.25, -0.06, p.z );
  pos[8]  = GzVectorSet( 0.25, -0.09, p.z );
  pos[9]  = GzVectorSet( 0.25, -0.12, p.z );
  pos[10] = GzVectorSet( 0.25, -0.15, p.z );

  pos[11] = GzVectorSet( 0.14, +0.18, p.z );
  pos[12] = GzVectorSet(-0.14, +0.18, p.z );
  pos[13] = GzVectorSet( 0.14, -0.18, p.z );
  pos[14] = GzVectorSet(-0.14, -0.18, p.z );

  pos[15] = GzVectorSet(-0.25, +0.12, p.z );
  pos[16] = GzVectorSet(-0.25, +0.00, p.z );
  pos[17] = GzVectorSet(-0.25, -0.12, p.z );
 
  dir[0] = 0;
  dir[1] = 0;
  dir[2] = 0;
  dir[3] = 0;
  dir[4] = 0;
  dir[5] = 0;
  dir[6] = 0;
  dir[7] = 0;
  dir[8] = 0;
  dir[9] = 0;
  dir[10] = 0;
  dir[11] = 90;
  dir[12] = 90;
  dir[13] = -90;
  dir[14] = -90;
  dir[15] = 180;
  dir[16] = 180;
  dir[17] = 180;

  sonarMinRange[0] = 0.04; sonarMaxRange[0] = 0.50;
  sonarMinRange[1] = 0.20; sonarMaxRange[1] = 1.00;
  sonarMinRange[2] = 0.04; sonarMaxRange[2] = 0.50;
  sonarMinRange[3] = 0.20; sonarMaxRange[3] = 1.00;
  sonarMinRange[4] = 0.04; sonarMaxRange[4] = 0.50;
  sonarMinRange[5] = 0.20; sonarMaxRange[5] = 1.00;
  sonarMinRange[6] = 0.04; sonarMaxRange[6] = 0.50;
  sonarMinRange[7] = 0.20; sonarMaxRange[7] = 1.00;
  sonarMinRange[8] = 0.04; sonarMaxRange[8] = 0.50;
  sonarMinRange[9] = 0.20; sonarMaxRange[9] = 1.00;

  sonarMinRange[10] = 0.04; sonarMaxRange[10] = 0.50;
  sonarMinRange[11] = 0.04; sonarMaxRange[11] = 0.50;
  sonarMinRange[12] = 0.04; sonarMaxRange[12] = 0.50;
  sonarMinRange[13] = 0.04; sonarMaxRange[13] = 0.50;
  sonarMinRange[14] = 0.04; sonarMaxRange[14] = 0.50;
  sonarMinRange[15] = 0.04; sonarMaxRange[15] = 0.50;
  sonarMinRange[16] = 0.04; sonarMaxRange[16] = 0.50;
  sonarMinRange[17] = 0.04; sonarMaxRange[17] = 0.50;

  this->sonarSensor = new RayProximity(this->world, this->chassis, 18);
  for (i = 0; i < 18; i++)
  {
    a = GzVectorSet(this->sonarMinRange[i] * cos(dir[i] * M_PI / 180),
                    this->sonarMinRange[i] * sin(dir[i] * M_PI / 180), 0.0);
    a = GzVectorAdd(pos[i], a); 

    b = GzVectorSet(this->sonarMaxRange[i] * cos(dir[i] * M_PI / 180),
                    this->sonarMaxRange[i] * sin(dir[i] * M_PI / 180), 0.0);
    b = GzVectorAdd(pos[i], b); 

    this->sonarSensor->SetRay(i, a, b);
  }
  
  return 0;
}


//////////////////////////////////////////////////////////////////////////////
// Initialize the model
int Protonek::Init( WorldFile *file, WorldFileNode *node )
{
  // Create position interface
  this->position_iface = gz_position_alloc();
  assert(this->position_iface);
  if (gz_position_create(this->position_iface, this->world->gz_server, this->GetId(),
                         "Protonek", (int) this, (int) this->parent) != 0)
    return -1;

  // Create power inteface
  this->power_iface = gz_power_alloc();
  if (gz_power_create(this->power_iface, this->world->gz_server, this->GetId(),
                      "Protonek", (int) this, (int) this->parent) != 0)
    return -1;

  // Create sonar inteface
  this->sonar_iface = gz_sonar_alloc();
  if (gz_sonar_create(this->sonar_iface, this->world->gz_server, this->GetId(),
                       "Protonek", (int) this, (int) this->parent) != 0)
    return -1;

  // Reset odometric pose
  this->odomPose[0] = 0.0;
  this->odomPose[1] = 0.0;
  this->odomPose[2] = 0.0;

  return 0;
}


//////////////////////////////////////////////////////////////////////////////
// Finalize the model
int Protonek::Fini()
{
  // Close sonar interface
  gz_sonar_destroy( this->sonar_iface );
  gz_sonar_free( this->sonar_iface );
  this->sonar_iface = NULL;

  // Close power interface
  gz_power_destroy( this->power_iface );
  gz_power_free( this->power_iface );
  this->power_iface = NULL;

  // Close position interface
  gz_position_destroy( this->position_iface );
  gz_position_free( this->position_iface );
  this->position_iface = NULL;

  return 0;
}


//////////////////////////////////////////////////////////////////////////////
// Update model
void Protonek::Update( double step )
{
  // Do nothing if paused
  if (step == 0)
    return;
  
  // Update the odometry (do this always for better accuracy)
  this->UpdateOdometry( step );

  // Otherwise, update periodically
  if (this->world->GetSimTime() - this->updateTime > this->updatePeriod)
  {
    this->updateTime = this->world->GetSimTime();

    // Get commands from the external interface
    this->GetPositionCmd();

    this->wheelJoints[0]->SetParam( dParamVel, 
                                    this->wheelSpeed[1] / this->wheelDiam * 2 );
    this->wheelJoints[1]->SetParam( dParamVel, 
                                    this->wheelSpeed[0] / this->wheelDiam * 2 );
    this->wheelJoints[2]->SetParam( dParamVel, 
                                    this->wheelSpeed[1] / this->wheelDiam * 2 );
    this->wheelJoints[3]->SetParam( dParamVel, 
                                    this->wheelSpeed[0] / this->wheelDiam * 2 );
    this->wheelJoints[4]->SetParam( dParamVel, 
                                    this->wheelSpeed[1] / this->wheelDiam * 2 );
    this->wheelJoints[5]->SetParam( dParamVel, 
                                    this->wheelSpeed[0] / this->wheelDiam * 2 );

    this->wheelJoints[0]->SetParam( dParamFMax, 10.0 );
    this->wheelJoints[1]->SetParam( dParamFMax, 10.0 );
    this->wheelJoints[2]->SetParam( dParamFMax, 10.0 );
    this->wheelJoints[3]->SetParam( dParamFMax, 10.0 );
    this->wheelJoints[4]->SetParam( dParamFMax, 10.0 );
    this->wheelJoints[5]->SetParam( dParamFMax, 10.0 );
  
    // Update the sonar sensor
    this->sonarSensor->Update();

    // Update the interface
    this->PutPositionData();
    this->PutPowerData();
    this->PutSonarData();
  }
  
  return;
}


//////////////////////////////////////////////////////////////////////////////
// Update the odometry
void Protonek::UpdateOdometry( double step )
{
  double wd, ws;
  double d1, d2;
  double dr, da;

  wd = this->wheelDiam;
  ws = this->wheelSep;

  // Average distance travelled by left and right wheels
  d1 = step * wd / 2 * 
    (wheelJoints[0]->GetAngleRate() + 
     wheelJoints[2]->GetAngleRate() +
     wheelJoints[4]->GetAngleRate()) / 3;
  d2 = step * wd / 2 * 
    (wheelJoints[1]->GetAngleRate() + 
     wheelJoints[3]->GetAngleRate() +
     wheelJoints[5]->GetAngleRate()) / 3;

  dr = (d1 + d2) / 2;
  da = (d2 - d1) / ws;
  
  // Compute odometric pose
  //this->odomPose[2] += da;
  this->odomPose[0] += dr * cos( this->odomPose[2] );
  this->odomPose[1] += dr * sin( this->odomPose[2] );
  this->odomPose[2] += da;

  // Compute odometric instantaneous velocity
  this->odomVel[0] = dr / step;
  this->odomVel[1] = 0.0;
  this->odomVel[2] = da / step;

  // Update the power discharge; this is probably completely bogus
  this->batteryLevel -= this->batteryCurve[0] * step;
  this->batteryLevel -= this->batteryCurve[1] * d1 * d1;
  this->batteryLevel -= this->batteryCurve[1] * d2 * d2;

  return;
}


//////////////////////////////////////////////////////////////////////////////
// Get commands from the external interface
void Protonek::GetPositionCmd()
{
  double vr, va;

  gz_position_lock(this->position_iface, 1);
  vr = this->position_iface->data->cmd_vel_pos[0];
  va = this->position_iface->data->cmd_vel_rot[2];
  gz_position_unlock(this->position_iface);
  
  this->wheelSpeed[0] = vr + va * this->wheelSep / 2;
  this->wheelSpeed[1] = vr - va * this->wheelSep / 2;
  
  return;
}


//////////////////////////////////////////////////////////////////////////////
// Update the data in the erinterface
void Protonek::PutPositionData()
{
  gz_position_lock(this->position_iface, 1);
    
  // Data timestamp
  this->position_iface->data->time = this->world->GetSimTime();

  this->position_iface->data->pos[0] = this->odomPose[0];
  this->position_iface->data->pos[1] = this->odomPose[1];
  this->position_iface->data->rot[2] = this->odomPose[2];

  this->position_iface->data->vel_pos[0] = this->odomVel[0];
  this->position_iface->data->vel_rot[2] = this->odomVel[2];

  gz_position_unlock(this->position_iface);
  
  return;
}


//////////////////////////////////////////////////////////////////////////////
// Update the data in the power interface
void Protonek::PutPowerData()
{
  gz_power_lock(this->power_iface, 1);
    
  this->power_iface->data->time = this->world->GetSimTime();
  this->power_iface->data->levels[0] = this->batteryLevel;

  gz_power_unlock(this->power_iface);
    
  return;
}


//////////////////////////////////////////////////////////////////////////////
// Update the data in the sonar interface
void Protonek::PutSonarData()
{
  int i;
  double r;

  gz_sonar_lock(this->sonar_iface, 1);
    
  // Data timestamp
  this->sonar_iface->data->time = this->world->GetSimTime();

  // Sonar count valid
  this->sonar_iface->data->sonar_count = 18;
  
  // Sonar's data
  for (i = 0; i < 18; i++)
  {
    r = Max(this->sonarSensor->GetRange(i), this->sonarMinRange[i]);
    r = Min(r, this->sonarMaxRange[i]);

    this->sonar_iface->data->sonar_pos[i][0] = pos[i].x;
    this->sonar_iface->data->sonar_pos[i][1] = pos[i].y;
    this->sonar_iface->data->sonar_pos[i][2] = pos[i].z;

    this->sonar_iface->data->sonar_rot[i][0] = 0;
    this->sonar_iface->data->sonar_rot[i][1] = 0;
    this->sonar_iface->data->sonar_rot[i][2] = dir[i] * M_PI / 180;

    this->sonar_iface->data->sonar_ranges[i] = r;
  }

  gz_sonar_unlock(this->sonar_iface);
    
  return;
}


