#include "includes.h"

//--------------------------------------------------------------
//--------------------------VERTEX------------------------------
//--------------------------------------------------------------

Vertex::Vertex (double xx, double yy, double zz) {
   Init(xx, yy, zz);
}

void Vertex::Init (double xx, double yy, double zz) {
   x = xx;
   y = yy;
   z = zz;
}


double Vertex::DistanceTo (Vertex &v) {
    double xd, yd, zd;

    xd = v.x - x;
    yd = v.y - y;
    zd = v.z - z;

    return sqrt (xd*xd + yd*yd + zd*zd);
}

//--------------------------------------------------------------
//--------------------------POLYGON-----------------------------
//--------------------------------------------------------------

Polygon::Polygon (void) {
    Vertex v(0, 0, 0);
    Init (v, v, v);
}

Polygon::Polygon (Vertex aa, Vertex bb, Vertex cc) {
    Init(aa, bb, cc);
}

int Polygon::IsValid (void) {
    if (a.DistanceTo(b) * 1000> MAX_EDGE_LENGTH) return 0;
    if (a.DistanceTo(c) * 1000> MAX_EDGE_LENGTH) return 0;
    if (b.DistanceTo(c) * 1000> MAX_EDGE_LENGTH) return 0;
    return 1;
}

int Polygon::IsValid (Vertex &aa, char va, Vertex &bb, char vb, Vertex &cc, char vc) {
    if (!va || !vb || !vc)                   return 0;
    if (aa.DistanceTo(bb) * 1000 > MAX_EDGE_LENGTH) return 0;
    if (aa.DistanceTo(cc) * 1000> MAX_EDGE_LENGTH) return 0;
    if (bb.DistanceTo(cc) * 1000> MAX_EDGE_LENGTH) return 0;
    return 1;
}

void Polygon::Init (Vertex aa, Vertex bb, Vertex cc) {
    a = aa; b = bb; c = cc;
}

void Polygon::WriteToFile (FILE *file_ptr) {
    fprintf (file_ptr, "%lf , %lf , %lf , %lf , %lf , %lf , %lf , %lf , %lf\n",
             a.x, a.z, a.y, b.x, b.z, b.y, c.x, c.z, c.y);
}

//--------------------------------------------------------------
//--------------------------SCAN_2D-----------------------------
//--------------------------------------------------------------

Scan2D::Scan2D (void) {
    cols = (int) (180.0 / STEP_LASER_ANGLE + 1.0);
    valid = new char [cols];
    col = new Vertex [cols];
}

Scan2D::Scan2D (const Scan2D &source) {
    int i;

    base = source.base;
    roll = source.roll;
    pitch = source.pitch;
    yaw = source.yaw;
    cols = source.cols;

    valid = new char [cols];
    col = new Vertex [cols];

    for (i=0; i<cols; i++) {
        valid[i] = source.valid[i];
        col[i] = source.col[i];
    }
}

Scan2D::~Scan2D (void) {
    delete [] col;
    delete [] valid;
    col = NULL;
    valid = NULL;
}

Scan2D& Scan2D::operator=(const Scan2D &source) {
    int i;
	
    if (this == &source) return *this;
    
    delete [] col;
    delete [] valid;
    col = NULL;
    valid = NULL;

    base = source.base;
    roll = source.roll;
    pitch = source.pitch;
    yaw = source.yaw;
    cols = source.cols;

    valid = new char [cols];
    col = new Vertex [cols];

    for (i=0; i<cols; i++) {
        valid[i] = source.valid[i];
	col[i] = source.col[i];
    }

    return *this;
}

int Scan2D::GetScan(PlayerClient &client, LaserProxy &laser, TruthProxy &truth, double ptz_angle) {
    int i;
    double ray, deg;
    double p0x, p0y;
    int valid_points = 0;

    logI ("Scan2D::GetScan: Collecting points at ptz_angle = %lf...\n", ptz_angle);
    
    get_truth_pos(client, truth, &base.x, &base.y, &base.z, &roll, &pitch, &yaw);

    pitch = deg2rad(ptz_angle);
    base.z += (double) LASER_TRUTH_Z_OFFSET / 1000.0;

    logSD ("Scan2D::GetScan: base.x = %lf, base.y = %lf, base.z = %lf\n", base.x, base.y, base.z);
    logSD ("Scan2D::GetScan: roll = %lf, pitch = %lf, yaw = %lf\n", rad2deg(roll), rad2deg(pitch), rad2deg(yaw));
    for (i = 0; i < cols; i++) {
        ray = laser[(int)(180.0 / STEP_LASER_ANGLE - i)];
        deg = -90 + i * STEP_LASER_ANGLE;

	p0x = ray * cos (pitch) * cos (deg2rad(deg));
	p0y = ray * sin (deg2rad(deg));
	
	col[i].x =   p0x * cos (yaw) + p0y * sin (yaw);
	col[i].y = - p0x * sin (yaw) + p0y * cos (yaw);
        col[i].z = base.z + ray * sin (pitch) * cos (deg2rad(deg));

	if (ray > (double) LASER_MAX_DISTANCE / 1000) 
		valid[i] = 0;
	else {       
		valid[i] = 1;  valid_points++; 
	}

	logSD ("Scan2D::GetScan: Yaw = %lf, Pitch = %lf, deg = %lf\n", yaw, pitch, deg);
	logSD ("Scan2D::GetScan: X = %lf, Y = %lf, Z = %lf, valid = %d\n", col[i].x, col[i].y, col[i].z, (int)valid[i]);
    }

    logD ("Scan2D::GetScan: Collected %d points (%d valid)\n", cols, valid_points);
    return valid_points;
}


int Scan2D::GetScan(PlayerClient &client, LaserProxy &laser, PositionProxy &position, double ptz_angle) {
    int i;
    double ray, deg;
    double p0x, p0y;
    int valid_points = 0;

    logI ("Scan2D::GetScan: Collecting points at ptz_angle = %lf...\n", ptz_angle);
    
    get_position(client, position, &base.x, &base.y, &yaw);

    pitch = deg2rad(ptz_angle);
    base.z = (double) LASER_TRUTH_Z_OFFSET / 1000.0;

    logSD ("Scan2D::GetScan: base.x = %lf, base.y = %lf, base.z = %lf\n", base.x, base.y, base.z);
    logSD ("Scan2D::GetScan: roll = %lf, pitch = %lf, yaw = %lf\n", rad2deg(roll), rad2deg(pitch), rad2deg(yaw));
    for (i = 0; i < cols; i++) {
        ray = laser[(int)(180.0 / STEP_LASER_ANGLE - i)];
        deg = -90 + i * STEP_LASER_ANGLE;

	p0x = ray * cos (pitch) * cos (deg2rad(deg));
	p0y = ray * sin (deg2rad(deg));
	
	col[i].x =   p0x * cos (yaw) + p0y * sin (yaw);
	col[i].y = - p0x * sin (yaw) + p0y * cos (yaw);
        col[i].z = base.z + ray * sin (pitch) * cos (deg2rad(deg));
	
	if (ray > (double) LASER_MAX_DISTANCE / 1000) 
		valid[i] = 0;
	else {       
		valid[i] = 1;  valid_points++; 
	}

	logSD ("Scan2D::GetScan: Yaw = %lf, Pitch = %lf, deg = %lf, L = %lf \n", yaw, pitch, deg, ray);
	logSD ("Scan2D::GetScan: X = %lf, Y = %lf, Z = %lf, valid = %d\n", col[i].x, col[i].y, col[i].z, (int)valid[i]);
    }

    logD ("Scan2D::GetScan: Collected %d points (%d valid)\n", cols, valid_points);
    return valid_points;
}


//--------------------------------------------------------------
//--------------------------SCAN_3D-----------------------------
//--------------------------------------------------------------

Scan3D::Scan3D (void) {
    rows = (int) ((MAX_TILT_ANGLE - MIN_TILT_ANGLE) / STEP_TILT_ANGLE + 1);
    cols = (int) (180 / STEP_LASER_ANGLE + 1);
    valid_points = 0;
    row = new Scan2D [rows];
    polygon_number = 0;
}

Scan3D::Scan3D (const Scan3D &source) {
    int i;

    rows = source.rows;
    cols = source.cols;
    valid_points = source.valid_points;
    polygon_number = source.polygon_number;
    polygons = source.polygons;

    row = new Scan2D [rows];

    for (i=0; i<rows; i++)
        row[i] = source.row[i];
	
}

Scan3D::~Scan3D (void) {
    delete [] row;
    row = NULL;
}

Scan3D& Scan3D::operator=(const Scan3D &source) {
    int i;

    if (this == &source) return *this;

    delete [] row;
    row = NULL;

    rows = source.rows;
    cols = source.cols;
    valid_points = source.valid_points;
    polygon_number = source.polygon_number;
    polygons = source.polygons;

    row = new Scan2D [rows];

    for (i=0; i<rows; i++)
        row[i] = source.row[i];

    return *this;
}

int Scan3D::GetScan(PlayerClient &client, LaserProxy &laser, TruthProxy &truth, PtzProxy &ptz) {
    double ptz_angle;
    int i;

    logI ("Scan3D::GetScan: Starting to collect 3D points!\n");
   
    if (wait_for_tilt(client, ptz, MIN_TILT_ANGLE) != 0) {
        logE("Scan3D::GetScan: Tilt value %lf could not be reached!\n", MIN_TILT_ANGLE);
        return -1;
    }
    
    for (i=0, ptz_angle = MIN_TILT_ANGLE; ptz_angle <= MAX_TILT_ANGLE; ptz_angle += STEP_TILT_ANGLE) {
        if (wait_for_tilt(client, ptz, ptz_angle) != 0) { 
            logE("Scan3D::GetScan: Tilt value %lf could not be reached!\n", ptz_angle);
            return -1;
        } 
        valid_points += row[i++].GetScan(client, laser, truth, ptz_angle);
    }

    logI ("Scan3D::GetScan: All %d 3D points (%d valid) successfully collected!\n", cols * rows, valid_points);

    return 0;
}

int Scan3D::GetScan(PlayerClient &client, LaserProxy &laser, PositionProxy &position, PtzProxy &ptz) {
    double ptz_angle;
    int i;

    logI ("Scan3D::GetScan: Starting to collect 3D points!\n");
   
    if (wait_for_tilt(client, ptz, MIN_TILT_ANGLE) != 0) {
        logE("Scan3D::GetScan: Tilt value %lf could not be reached!\n", MIN_TILT_ANGLE);
        return -1;
    }
    
    for (i=0, ptz_angle = MIN_TILT_ANGLE; ptz_angle <= MAX_TILT_ANGLE; ptz_angle += STEP_TILT_ANGLE) {
        if (wait_for_tilt(client, ptz, ptz_angle) != 0) { 
            logE("Scan3D::GetScan: Tilt value %lf could not be reached!\n", ptz_angle);
            return -1;
        } 
        valid_points += row[i++].GetScan(client, laser, position, ptz_angle);
    }

    logI ("Scan3D::GetScan: All %d 3D points (%d valid) successfully collected!\n", cols * rows, valid_points);

    return 0;
}




int Scan3D::CalculatePolygons (void) {
    int i, j;
    int ile = 0;
    Vertex A, B, C, D;
    char va, vb, vc, vd;
    char vABC, vABD, vACD, vBCD;

    for (i=0; i < rows - 1; i++)
        for (j=0; j < cols - 1; j++) {

	    vABC = 0; vABD = 0; vACD = 0; vBCD = 0;
		
	    A = row[i  ].col[j  ]; va = row[i  ].valid[j  ];
	    B = row[i  ].col[j+1]; vb = row[i  ].valid[j+1];
	    C = row[i+1].col[j  ]; vc = row[i+1].valid[j  ];
	    D = row[i+1].col[j+1]; vd = row[i+1].valid[j+1];
		
            if (Polygon::IsValid(A, va, B, vb, C, vc)) vABC = 1;
            if (Polygon::IsValid(A, va, B, vb, D, vd)) vABD = 1;
            if (Polygon::IsValid(A, va, C, vc, D, vd)) vACD = 1;
            if (Polygon::IsValid(B, vb, C, vc, D, vd)) vBCD = 1;

	    //0
	    if (!vABC && !vABD && !vACD && !vBCD)  continue; 
	    
	    //1
	    if (vABC && !vABD && !vACD && !vBCD) { polygons.push_back(Polygon(A,B,C)); ile++; continue; }
	    if (!vABC && vABD && !vACD && !vBCD) { polygons.push_back(Polygon(A,B,D)); ile++; continue; }
	    if (!vABC && !vABD && vACD && !vBCD) { polygons.push_back(Polygon(A,D,C)); ile++; continue; }
	    if (!vABC && !vABD && !vACD && vBCD) { polygons.push_back(Polygon(B,D,C)); ile++; continue; }

	    //2
	    if (vABC && vABD && !vACD && !vBCD) { polygons.push_back(Polygon(A,B,C)); ile++; continue; }
	    if (vABC && !vABD && vACD && !vBCD) { polygons.push_back(Polygon(A,B,C)); ile++; continue; }
	    if (vABC && !vABD && !vACD && vBCD) { polygons.push_back(Polygon(A,B,C)); ile++; 
		                                  polygons.push_back(Polygon(B,D,C)); ile++; continue; }
	    if (!vABC && vABD && vACD && !vBCD) { polygons.push_back(Polygon(A,B,D)); ile++; 
		                                  polygons.push_back(Polygon(A,D,C)); ile++; continue; }
	    if (!vABC && vABD && !vACD && vBCD) { polygons.push_back(Polygon(A,B,D)); ile++; continue; }
	    if (!vABC && !vABD && vACD && vBCD) { polygons.push_back(Polygon(A,D,C)); ile++; continue; }
	    
	    //3
	    if (vABC && vABD && vACD && !vBCD) { polygons.push_back(Polygon(A,B,D)); ile++; 
		                                 polygons.push_back(Polygon(A,D,C)); ile++; continue; }
	    if (vABC && vABD && !vACD && vBCD) { polygons.push_back(Polygon(A,B,C)); ile++; 
		                                 polygons.push_back(Polygon(B,D,C)); ile++; continue; }
	    if (vABC && !vABD && vACD && vBCD) { polygons.push_back(Polygon(A,B,C)); ile++; 
		                                 polygons.push_back(Polygon(B,D,C)); ile++; continue; }
	    if (!vABC && vABD && vACD && vBCD) { polygons.push_back(Polygon(A,B,D)); ile++; 
		                                 polygons.push_back(Polygon(A,D,C)); ile++; continue; }

            //4
	    if (A.DistanceTo(D) < B.DistanceTo(C)) { polygons.push_back(Polygon(A,B,D)); ile++; 
		                                     polygons.push_back(Polygon(A,D,C)); ile++; continue; }

	
            polygons.push_back(Polygon(A,B,C)); ile++; 
	    polygons.push_back(Polygon(B,D,C)); ile++; 
	}
    polygon_number = ile;
    return ile;
}

//--------------------------------------------------------------
//---------------------------SCENE------------------------------
//--------------------------------------------------------------


Scene::Scene (void) { 
    valid_points = 0;
    polygon_number = 0;
}

Scene::~Scene (void) { }

void Scene::AddScan (PlayerClient &client, LaserProxy &laser, TruthProxy &truth, PtzProxy &ptz) {
    Scan3D scan;

    scan.GetScan(client, laser, truth, ptz);
    valid_points += scan.valid_points;
    
    scans.push_back(scan);
}

void Scene::AddScan (PlayerClient &client, LaserProxy &laser, PositionProxy &position, PtzProxy &ptz) {
    Scan3D scan;

    scan.GetScan(client, laser, position, ptz);
    valid_points += scan.valid_points;
    
    scans.push_back(scan);
}



void Scene::WriteCones (char *filename) {
    FILE *file_ptr;
    int i, j, s;
    double tri_size = 0.005;
    Vertex a, b, c, d;
    Polygon A, B, C, D;

    if (scans.size() == 0) {
        logW("Scene::WriteCones: Nothing to write!\n");
	return;
    }
    
    file_ptr = fopen (filename, "w");

    logI ("Scene::WriteCones: Writing scene to file \"%s\"\n", filename);
    if (USE_INVALID_POINTS)
    	fprintf (file_ptr, "beginmodel %d\n", scans.size() * scans[0].rows * scans[0].cols * 4);
    else
        fprintf (file_ptr, "beginmodel %d\n", valid_points*4);
    
    for (s=0; s < scans.size(); s++)
    for (i=0; i < scans[s].rows; i++)
        for (j=0; j < scans[s].cols; j++) {
	    
	    if (scans[s].row[i].valid[j] == 0 && !USE_INVALID_POINTS) continue;

	    a.Init (scans[s].row[i].col[j].x - tri_size, scans[s].row[i].col[j].y - tri_size, 
	            scans[s].row[i].col[j].z - tri_size);
	    
	    b.Init (scans[s].row[i].col[j].x - tri_size, scans[s].row[i].col[j].y + tri_size, 
		    scans[s].row[i].col[j].z - tri_size);
	    
	    c.Init (scans[s].row[i].col[j].x + tri_size, scans[s].row[i].col[j].y, 
		    scans[s].row[i].col[j].z - tri_size);
	    
	    d.Init (scans[s].row[i].col[j].x, scans[s].row[i].col[j].y, scans[s].row[i].col[j].z + tri_size);

	    A.Init (a, b, c); 
	    B.Init (a, b, d); 
	    C.Init (a, c, d); 
	    D.Init (b, c, d);
	     
	    A.WriteToFile(file_ptr);
	    B.WriteToFile(file_ptr);
	    C.WriteToFile(file_ptr);
	    D.WriteToFile(file_ptr);

	}
    fclose (file_ptr);   
}

void Scene::CalculatePolygons (void) {
    int i;

    for (i=0; i < scans.size(); i++)
	if (scans[i].polygon_number == 0)
            polygon_number += scans[i].CalculatePolygons();
}

void Scene::WritePolygons (char *filename) {
    FILE *file_ptr;
    int i, j;

    CalculatePolygons();
    
    if (polygon_number == 0) {
        logW("Scene::WritePolygons: Nothing to write!\n");
        return;
    }

    file_ptr = fopen (filename, "w");

    logI ("Scene::WritePolygons: Writing scene to file \"%s\"\n", filename);
    fprintf (file_ptr, "beginmodel %d\n", polygon_number);

    for (i=0; i < scans.size(); i++)
        for (j=0; j < scans[i].polygon_number; j++) 
	    scans[i].polygons[j].WriteToFile(file_ptr);

    fclose (file_ptr);   
}
