/*
 *  Gazebo - Outdoor Multi-Robot Simulator
 *  Copyright (C) 2003  
 *     Nate Koenig & Andrew Howard
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/* Desc: Model for a Tilt_SickLMS200
 * Author: Andrew Howard
 * Date: 8 May 2003
 * CVS: $Id: Tilt_SickLMS200.cc,v 1.3 2006/04/19 20:14:16 kniewiar Exp $
 */

/// @addtogroup models
/// @{ 
/** @defgroup Tilt_SickLMS200 Sick LMS 200

@htmlinclude Tilt_SickLMS200_view.html

The Tilt_SickLMS200 model simulates the ubiquitous SICK LMS200 scanning
laser range-finder.

@par libgazebo interfaces

This model supports the @ref laser and @ref fiducial interfaces.

@par Player drivers

Range and intensity data is available through the %gz_laser driver.
Fiducial information (ID, range, bearing and orientation) is available
through the %gz_fiducial driver.  At this time, only the @ref
SimpleSolid model will return a non-zero fiducial value.


@par Attributes

The following attributes are supported.

@htmlinclude default_attr_include.html

- rangeCount (integer)
  - The number of range simulated readings to generate.
  - Default: 181

- rayCount (integer)
  - The number of rays to test.  If rangeCount > rayCount, the
    additional readings will be interpolated.  Reducing the number of
    rays is a good way to save CPU cycles (at the expense of
    simulation fidelity).
  - Default: 91
  
- rangeMax (float, m)
  - Maximum laser range.
  - Default: 8.192
  
- rangeMin (float, m)
  - Minimum laser range.
  - Default: 0.20
  
- scanRate (float, Hz)
  - The scanning rate, or scans-per-second.
  - Default: 10


@par Bodies

The following bodies are created by this model.

@htmlinclude default_body_include.html

@par Example

@verbatim
<model:Tilt_SickLMS200>
  <id>laser1</id>
  <xyz>0.15 0 0.20</xyz>
  <rangeCount>181</rangeCount>
  <scanRate>10</scanRate>
</model:Tilt_SickLMS200>
@endverbatim

@par Views

@htmlinclude Tilt_SickLMS200_more_views.html

@par Authors

Nate Koenig, Andrew Howard

*/
/// @}

#define LOG 0
#define LMS_HEIGHT 0.05

  
#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <time.h>

#include <gazebo.h>
#include <gazebo/World.hh>
#include <gazebo/WorldFile.hh>
#include <gazebo/ModelFactory.hh>
#include <gazebo/RayProximity.hh>
#include <gazebo/BoxGeom.hh>
#include <gazebo/CylinderGeom.hh>
#include <gazebo/HingeJoint.hh>
#include "Tilt_SickLMS200.hh"



//////////////////////////////////////////////////////////////////////////////
// Register this model
GZ_REGISTER_PLUGIN("Tilt_SickLMS200", Tilt_SickLMS200);


//////////////////////////////////////////////////////////////////////////////
// Constructor
Tilt_SickLMS200::Tilt_SickLMS200( World *world )
    : Model( world )
{
  LogMsg("Constructor Start\n","msg.log");
  this->LMS = NULL;
  this->Servo = NULL;
  this->sensor = NULL;
  this->tiltJoint = NULL;
  
  this->ptzIface = gz_ptz_alloc();  

  LogMsg("Constructor End\n","msg.log");
  return;
}


//////////////////////////////////////////////////////////////////////////////
// Destructor
Tilt_SickLMS200::~Tilt_SickLMS200()
{
  LogMsg("Destructor Start\n","msg.log");
  if (this->LMS)
    delete this->LMS;
  if (this->Servo)
    delete this->Servo;
  if (this->tiltJoint)
    delete this->tiltJoint;
  if (this->sensor)
    delete this->sensor;

  this->LMS = NULL;
  this->Servo = NULL;
  this->tiltJoint = NULL;
  this->sensor = NULL;

  this->ptzIface = NULL;
  
  LogMsg("Destructor End\n","msg.log");
  return;
}


//////////////////////////////////////////////////////////////////////////////
// Load the model
int Tilt_SickLMS200::Load( WorldFile *file, WorldFileNode *node )
{
  LogMsg("Load Start\n","msg.log");
  
  // Get the time between ptz updates
  this->ptzUpdatePeriod = 0.100;
  this->ptzUpdateTime = -this->ptzUpdatePeriod;

  // Gain term on PTZ controller
  this->motionGain = node->GetDouble( "motionGain", 2.0 );
  
  // Number of simulated  rays
  this->rayCount = node->GetInt("rayCount", 91);

  // Number of range readings
  this->rangeCount = node->GetInt("rangeCount", 181);

  this->laserPeriod = 1.0 / (node->GetDouble("scanRate", 10) + 1e-6);
  this->laserMaxRange = node->GetLength("rangeMax", 8.192);
  this->laserMinRange = node->GetLength("rangeMin", 0.20);
  this->laserMinAngle = node->GetAngle("minAngle", -M_PI * 0.5);
  this->laserMaxAngle = node->GetAngle("maxAngle", +M_PI * 0.5);

  // Create the ODE objects
  if (this->OdeLoad(file, node) != 0)
    return -1;

  // Create ray sensor
  if (this->RayLoad(file, node) != 0)
    return -1;
 
  LogMsg("Load End\n","msg.log");
  return 0;
}


//////////////////////////////////////////////////////////////////////////////
// Load ODE objects
int Tilt_SickLMS200::OdeLoad( WorldFile *file, WorldFileNode *node )
{
  double oz, length, width, height, mass;
  double Servo_base_dim[3], Servo_col_dim[3];
  double Servo_base_mass, Servo_col_mass;
  Geom *geom;
  
  LogMsg("OdeLoad Start\n","msg.log");
  
  length = 0.15;
  width = 0.15;
  height = 0.20;
  mass = 7.0;

  //Servo column
  Servo_col_dim[0] = 0.05; //głębokość
  Servo_col_dim[1] = 0.01; //szerokość
  Servo_col_dim[2] = LMS_HEIGHT + height / 2; //wysokość
  Servo_col_mass   = 0.5;
  
  //Servo base
  Servo_base_dim[0] = 0.15;                          //głębokość
  Servo_base_dim[1] = width + 2 * Servo_col_dim[1];  //szerokość
  Servo_base_dim[2] = 0.01;                          //wysokość
  Servo_base_mass   = 1.0;
  
  oz = height / 2;

  //Servo body  
  this->Servo = new Body( this->world );
  
  geom = new BoxGeom (this->Servo, this->modelSpaceId, Servo_base_dim[0], Servo_base_dim[1], Servo_base_dim[2]);
  geom->SetRelativePosition(GzVectorSet(0, 0, Servo_base_dim[2] / 2));
  geom->SetMass (Servo_base_mass);
  geom->SetColor( GzColor(0.3, 0.3, 0.3) );

  geom =  new BoxGeom (this->Servo, this->modelSpaceId, Servo_col_dim[0], Servo_col_dim[1], Servo_col_dim[2]);
  geom->SetRelativePosition(GzVectorSet(
			  0, 
			  Servo_base_dim[1] / 2 - Servo_col_dim[1] / 2, 
			  Servo_base_dim[2] + Servo_col_dim[2] / 2));
  geom->SetMass (Servo_col_mass);
  geom->SetColor( GzColor(0.3, 0.3, 0.3) );
	
  geom =  new BoxGeom (this->Servo, this->modelSpaceId, Servo_col_dim[0], Servo_col_dim[1], Servo_col_dim[2]);
  geom->SetRelativePosition(GzVectorSet(
			  0, 
			  0 - Servo_base_dim[1] / 2 + Servo_col_dim[1] / 2, 
			  Servo_base_dim[2] + Servo_col_dim[2] / 2));
  geom->SetMass (Servo_col_mass);
  geom->SetColor( GzColor(0.3, 0.3, 0.3) );
  
  this->AddBody( this->Servo, true);


  // LMS200 body
  this->LMS = new Body( this->world );
  
  geom = new BoxGeom( this->LMS, this->modelSpaceId, length, width, 0.2 * height );
  geom->SetRelativePosition(GzVectorSet(0, 0, LMS_HEIGHT + oz + (-0.5 + 0.2 / 2) * height) );
  geom->SetMass( mass / 3 );
  geom->SetColor( GzColor(0, 0, 1) );
  
  geom = new BoxGeom( this->LMS, this->modelSpaceId, length, width, 0.4 * height );
  geom->SetRelativePosition(GzVectorSet(0, 0, LMS_HEIGHT + oz + (+0.5 - 0.4 / 2) * height));
  geom->SetMass( mass / 3 );
  geom->SetColor( GzColor(0, 0, 1) );
  
  geom = new BoxGeom( this->LMS, this->modelSpaceId, length / 2, width, 0.9 * height );
  geom->SetRelativePosition(GzVectorSet( -0.25 * length, 0, LMS_HEIGHT + oz));
  geom->SetMass( mass / 3 );
  geom->SetColor( GzColor(0, 0, 1) );
  
  geom = new CylinderGeom( this->LMS, this->modelSpaceId, length / 2, 0.9 * height );
  geom->SetRelativePosition( GzVectorSet(0, 0, LMS_HEIGHT + oz) );
  geom->SetColor( GzColor(0, 0, 0) );

  this->AddBody( this->LMS);


  //tilt joint
  this->tiltJoint = new HingeJoint( this->world->worldId );
  this->tiltJoint->Attach(this->LMS, this->Servo);
  this->tiltJoint->SetAnchor( 0, 0, LMS_HEIGHT + height / 2);
  this->tiltJoint->SetAxis(0,1,0);
  this->tiltJoint->SetParam( dParamFMax, 1000 );

  this->LMS->SetFiniteRotationMode(1);

  LogMsg("OdeLoad End\n","msg.log");
  return 0;
}


//////////////////////////////////////////////////////////////////////////////
// Load rays
int Tilt_SickLMS200::RayLoad( WorldFile *file, WorldFileNode *node )
{
  int i;
  double angle;
  GzVector org, axis, pos, qos;

  //GzQuatern rot;

  LogMsg("RayLoad Start\n","msg.log");

  // Create the sensor
  this->sensor = new RayProximity(this->world, this->LMS, this->rayCount);
  
  // Origin for laser rays
  org = GzVectorSet(0, 0, 0.09 + LMS_HEIGHT);
  
  // Create the laser rays
  for (i = 0; i < this->rayCount; i++)
  {
    angle = i * (this->laserMaxAngle - this->laserMinAngle) / (this->rayCount - 1) + this->laserMinAngle;
    axis = GzVectorSet(cos(angle), sin(angle), 0);
        
    // Origin of ray
    pos = GzVectorMul(this->laserMinRange, axis);
    pos = GzVectorAdd(pos, org);

    // End point of ray (unit length)
    qos = GzVectorMul(this->laserMaxRange, axis);
    qos = GzVectorAdd(qos, org);

    // Set ray endpoints
    this->sensor->SetRay(i, pos, qos);
  }

  LogMsg("RayLoad End\n","msg.log");
  return 0;
}


//////////////////////////////////////////////////////////////////////////////
// Initialize the model
int Tilt_SickLMS200::Init( WorldFile *file, WorldFileNode *node )
{

  LogMsg("Init Start\n","msg.log");
  // Create laser interface
  this->laser_iface = gz_laser_alloc();
  assert(this->laser_iface);
  if (gz_laser_create(this->laser_iface, this->world->gz_server, this->GetId(),
                      "Tilt_SickLMS200", (int) this, (int) this->parent) != 0)
    return -1;

  // PTZ interface
  assert(this->ptzIface);
  if (gz_ptz_create(this->ptzIface, this->world->gz_server, this->GetId(),
                    "Tilt_SickLMS200", (int) this, (int) this->parent) !=0)
    return -1;

  // Initialize PTZ interface to sensible values
  this->ptzIface->data->cmd_zoom = M_PI;


  this->laserTime = this->world->GetSimTime();
  
  LogMsg("Init End\n","msg.log");
  return 0;
}



//////////////////////////////////////////////////////////////////////////////
// Finalize the model
int Tilt_SickLMS200::Fini()
{
  LogMsg("Fini Start\n","msg.log");
    
  gz_laser_destroy( this->laser_iface );
  gz_laser_free( this->laser_iface );
  this->laser_iface = NULL;

  gz_ptz_destroy(this->ptzIface);
  gz_ptz_free(this->ptzIface);
  this->ptzIface = NULL;


  LogMsg("Fini End\n","msg.log");
  return 0;
}


//////////////////////////////////////////////////////////////////////////////
// Update the model
void Tilt_SickLMS200::Update( double step )
{
  LogMsg("Update Start\n","msg.log");

  double tilt;

  if (this->world->GetSimTime() - this->ptzUpdateTime >= this->ptzUpdatePeriod)
  {
    step = this->world->GetSimTime() - this->ptzUpdateTime;

    this->ptzUpdateTime = this->world->GetSimTime();

    // Get new commands
    this->GetPtzCmd();

    // Apply joint limits to commanded pan/tilt angles
    if (this->cmdTilt > M_PI*0.5)
      this->cmdTilt = M_PI*0.5;
    else if (this->cmdTilt < -M_PI*0.5)
      this->cmdTilt = -M_PI*0.5;


    // Set the pan and tilt motors; can't set angles so track cmds with
    // a proportional control
    tilt = this->cmdTilt - this->tiltJoint->GetAngle();
    if (tilt < 0 ) tilt -= 0.1/180*M_PI; //these lines are workaround 
    if (tilt > 0 ) tilt += 0.1/180*M_PI; //for player ptz interface bug
    this->tiltJoint->SetParam( dParamVel, this->motionGain * tilt);

    // Set the zoom using a proportional controller
//    this->zoom += (this->cmdZoom - this->zoom) * this->zoomGain * step;

    // Update the ptz interface
    this->PutPtzData();
  }


  // Update the laser periodically
  if (this->world->GetSimTime() - this->laserTime < this->laserPeriod)
    return;
  this->laserTime = this->world->GetSimTime();
  IFaceGetCmd();

  // Update the ray sensor
  this->sensor->Update();
  
  // Update the mmap interface with the new data
  this->PutLaserData();
  
  LogMsg("Update End\n","msg.log");
  return;
}


//////////////////////////////////////////////////////////////////////////////
// Send PTZ data
void Tilt_SickLMS200::PutPtzData()
{
  char msg[100];
  // Update the PTZ interface
  gz_ptz_lock(this->ptzIface, 1);
  this->ptzIface->data->time = this->world->GetSimTime();
  this->ptzIface->data->tilt = this->tiltJoint->GetAngle();
  sprintf (msg, "tilt = %lf\n", this->ptzIface->data->tilt);
  LogMsg(msg, "msg.log");
  gz_ptz_unlock(this->ptzIface);
  gz_ptz_post(this->ptzIface);
  
  return;
}



//////////////////////////////////////////////////////////////////////////////
// Update the data in the interface
void Tilt_SickLMS200::PutLaserData()
{
  int i, ja, jb;
  double ra, rb, r, b;
  int v;

  LogMsg("PutLaserData Start\n","msg.log");
  gz_laser_lock(this->laser_iface, 1);

  // Data timestamp
  this->laser_iface->data->time = this->world->GetSimTime();
    
  // Read out the laser range data
  this->laser_iface->data->min_angle = this->laserMinAngle;
  this->laser_iface->data->max_angle = this->laserMaxAngle;
  this->laser_iface->data->res_angle = (this->laserMaxAngle - this->laserMinAngle) / (this->rangeCount - 1);
  this->laser_iface->data->max_range = this->laserMaxRange;
  this->laser_iface->data->range_count = this->rangeCount;
  

  assert( this->laser_iface->data->range_count < GZ_LASER_MAX_RANGES );

  // Interpolate the range readings from the rays
  for (i = 0; i < this->rangeCount; i++)
  {
    b = (double) i * (this->rayCount - 1) / (this->rangeCount - 1);
    ja = (int) floor(b);
    jb = Min(ja + 1, this->rayCount - 1);    
    b = b - floor(b);

    assert(ja >= 0 && ja < this->rayCount);
    assert(jb >= 0 && jb < this->rayCount);

    //ra = Min(this->sensor->GetRange(ja) + this->laserMinRange, this->laserMaxRange);
    //rb = Min(this->sensor->GetRange(jb) + this->laserMinRange, this->laserMaxRange);
    ra = Min(this->sensor->GetRange(ja) , this->laserMaxRange);
    rb = Min(this->sensor->GetRange(jb) , this->laserMaxRange);

    //printf("%d %d %d %f \n","msg.log", i, ja, jb, ra);

    // Range is linear interpolation if values are close,
    // and min if they are very different
    if (fabs(ra - rb) < 0.10)
      r = (1 - b) * ra + b * rb;
    else
      r = Min(ra, rb);

    // Intensity is either-or
    v = (int) this->sensor->GetRetro(ja) || (int) this->sensor->GetRetro(jb);

    assert(i < GZ_LASER_MAX_RANGES);
    this->laser_iface->data->ranges[i] =  r + this->laserMinRange;
    this->laser_iface->data->intensity[i] = v;
  }

  gz_laser_unlock(this->laser_iface);

  // New data is available
  gz_laser_post(this->laser_iface);
  
  LogMsg("PutLaserData End\n","msg.log");
  return;
}


//////////////////////////////////////////////////////////////////////////////
// Get commands
void Tilt_SickLMS200::GetPtzCmd()
{
  gz_ptz_lock(this->ptzIface, 1);

  this->cmdTilt = this->ptzIface->data->cmd_tilt;
//  this->cmdZoom = this->hfov / this->ptzIface->data->cmd_zoom;

  gz_ptz_unlock(this->ptzIface);
  gz_ptz_post(this->ptzIface);

  return;
}


//////////////////////////////////////////////////////////////////////////////
// Get command from the interface
bool Tilt_SickLMS200::IFaceGetCmd()
{
  bool cmd_new;
  
  LogMsg("IFaceGetCmd Start\n","msg.log");
  gz_laser_lock(this->laser_iface,1);

  if (this->laser_iface->data->cmd_new_angle || this->laser_iface->data->cmd_new_length)
    {
      cmd_new = true;
      if (this->laser_iface->data->cmd_new_angle == 1)
	this->laser_iface->data->cmd_new_angle = 0;
      else if (this->laser_iface->data->cmd_new_length == 1)
	this->laser_iface->data->cmd_new_length = 0;

      this->laserMaxRange = this->laser_iface->data->cmd_max_range;
      this->laserMinAngle = this->laser_iface->data->cmd_min_angle;
      this->laserMaxAngle = this->laser_iface->data->cmd_max_angle;
      this->rangeCount = this->laser_iface->data->cmd_range_count;
    }
  else
    cmd_new = false;

  gz_laser_unlock(this->laser_iface);

  LogMsg("IFaceGetCmd End\n","msg.log");
  return cmd_new;
}


void Tilt_SickLMS200::LogMsg(char *msg, char *file_name)
{
  FILE *file;
  time_t rawtime;
  char current_time[30];
  
  if (LOG == 0) return;
  
  time ( &rawtime );
  ctime_r (&rawtime, current_time);
  
  for (int i = 0; i<30; i++)
    if (current_time[i] == '\n') current_time[i] = '\0';
  
  file = fopen(file_name, "a");
 
  fprintf (file, "%s : ", current_time );
  fprintf (file, msg);
  
  fclose(file);
}
